import numpy as np
from pathlib import Path

# -------------------------------------------------------------------
# 1. Parametry LFSR
# -------------------------------------------------------------------
LFSR_BITS = 40
# Wybierz tu tapy zgodne z wybranym wielomianem (indeksy od 0 = LSB do 39 = MSB)
# Przykładowo dla wielomianu x^40 + x^7 + x^2 + x + 1:
TAPS = [39, 6, 1, 0]

# Początkowy stan LFSR (nie powinien być zerowy)
# Możesz tu wstawić dowolne 40-bitowe nasienie
lfsr_state = (1 << LFSR_BITS) - 1  # wszystkie bity = 1

# -------------------------------------------------------------------
# 2. Wczytaj surowe bajty i rozwiń na bity
# -------------------------------------------------------------------
in_path = Path("rng_bin_13MB.bin")   # Twój plik źródłowy
with in_path.open("rb") as f:
    raw = f.read()

# Konwersja na tablicę uint8 i potem na listę bitów (MSB first)
bytes_arr = np.frombuffer(raw, dtype=np.uint8)
src_bits = []
for b in bytes_arr:
    # MSB-->LSB: jeśli wolisz odwrotnie, zmień range na (0,8)
    for i in range(7, -1, -1):
        src_bits.append((b >> i) & 1)

# -------------------------------------------------------------------
# 3. Nieregularne próbkowanie LFSR (clocking)
# -------------------------------------------------------------------
out_bits = []
for bit in src_bits:
    # clocking: jeśli źródłowy bit = 1, wykonaj krok LFSR i pobierz wyjście
    if bit == 1:
        # oblicz nowy bit jako XOR pozycjo TAPS
        new_bit = 0
        for t in TAPS:
            new_bit ^= (lfsr_state >> t) & 1
        # shift w prawo o 1, wstaw new_bit na MSB (bit 39)
        lfsr_state = (lfsr_state >> 1) | (new_bit << (LFSR_BITS - 1))
        # pobierz LSB rejestru jako bit wyjściowy
        out_bits.append(lfsr_state & 1)
    # w przeciwnym razie „czekaj” (brak kroku, brak wyjścia)

# -------------------------------------------------------------------
# 4. Grupowanie bitów w bajty i zapis
# -------------------------------------------------------------------
# ile pełnych bajtów mamy?
n_bytes = len(out_bits) // 8

out_bytes = bytearray()
for i in range(n_bytes):
    byte = 0
    for b in out_bits[i*8:(i+1)*8]:
        byte = (byte << 1) | b
    out_bytes.append(byte)

out_path = Path("postprocessed.bin")
with out_path.open("wb") as f:
    f.write(out_bytes)

print(f"Wygenerowano {len(out_bytes)} bajtów post-processed i zapisano do {out_path}")
