import numpy as np
from pathlib import Path

# -------------------------------------------------------------------
# 1. Parametry LFSR
# -------------------------------------------------------------------
LFSR_BITS = 40
TAPS = [39, 6, 1, 0]  # Tapy dla wielomianu x^40 + x^7 + x^2 + x + 1
lfsr_state = (1 << LFSR_BITS) - 1  # Początkowy stan LFSR (wszystkie bity = 1)

# -------------------------------------------------------------------
# 2. Funkcja do wczytywania danych o logarytmicznych stopach zwrotu z pliku binarnego
# -------------------------------------------------------------------
def load_btc_returns(file_path):
    with open(file_path, 'rb') as f:
        byte_data = f.read()
    # Zamiana na tablicę float32
    btc_returns = np.frombuffer(byte_data, dtype=np.float32)
    return btc_returns

# -------------------------------------------------------------------
# 3. Nieregularne próbkowanie LFSR (clocking) na podstawie danych BTC
# -------------------------------------------------------------------
def lfsr_clocking_with_btc_data(btc_returns, lfsr_state, taps, lfsr_bits=40, target_size_bytes=13 * 1024 * 1024):
    out_bits = []
    step_count = 0  # Licznik kroków LFSR
    target_size_bits = target_size_bytes * 8  # Konwersja docelowego rozmiaru na bity
    
    i = 0  # Licznik iteracji po danych
    while len(out_bits) < target_size_bits:
        log_return = btc_returns[i % len(btc_returns)]  # Dla danych, które mają się powtarzać
        
        # Zastosowanie LSB z logarytmicznej stopy zwrotu (log_return) jako entropii
        lsb_bit = 1 if log_return > 0 else 0  # Zmienność (jeśli zwrot dodatni, LSB = 1, w przeciwnym razie 0)
        
        # Próbkowanie LFSR
        if lsb_bit == 1:  # Kiedy LSB = 1, wykonujemy krok LFSR
            new_bit = 0
            for t in taps:
                new_bit ^= (lfsr_state >> t) & 1
            lfsr_state = (lfsr_state >> 1) | (new_bit << (lfsr_bits - 1))
            out_bits.append(lfsr_state & 1)
            step_count += 1  # Inkrementacja licznika kroków LFSR
        
        i += 1  # Przechodzimy do następnego odczytu z danych BTC
    
    return out_bits, step_count

# -------------------------------------------------------------------
# 4. Grupowanie bitów w bajty i zapis do pliku
# -------------------------------------------------------------------
def save_bits_to_file(bits, output_filename="postprocessed.bin"):
    # Grupowanie bitów w bajty
    n_bytes = len(bits) // 8
    out_bytes = bytearray()
    for i in range(n_bytes):
        byte = 0
        for b in bits[i*8:(i+1)*8]:
            byte = (byte << 1) | b
        out_bytes.append(byte)

    # Zapis do pliku
    out_path = Path(output_filename)
    with out_path.open("wb") as f:
        f.write(out_bytes)
    print(f"Wygenerowano {len(out_bytes)} bajtów post-processed i zapisano do {out_path}")

# -------------------------------------------------------------------
# 5. MAIN - Odczyt danych i przetwarzanie
# -------------------------------------------------------------------
if __name__ == "__main__":
    # Wczytanie danych z pliku binarnego z logarytmicznymi stopami zwrotu
    btc_returns = load_btc_returns('btc_returns_13MB.bin')  # Plik zawierający logarytmiczne stopy zwrotu

    # Przetwarzanie danych przy użyciu LFSR
    out_bits, step_count = lfsr_clocking_with_btc_data(btc_returns, lfsr_state, TAPS, target_size_bytes=13 * 1024 * 1024)

    # Zapisanie wygenerowanych bitów do pliku binarnego
    save_bits_to_file(out_bits, "postprocessed.bin")

    # Wyświetlenie liczby wykonanych kroków LFSR
    print(f"Liczba wykonanych kroków LFSR: {step_count}")
